# $XTermId: VTparse.def,v 1.31 2008/01/27 17:08:54 tom Exp $
#
# vile:confmode rs=lf
#
# List of symbols that need to be defined for VTparse.h.  If you need to
# change any of the CASE_ macros, make the change here and rerun the command
# shown in VTparse.h.
#

CASE_GROUND_STATE
CASE_IGNORE
CASE_BELL
CASE_BS
CASE_CR
CASE_ESC
CASE_VMOT
CASE_TAB
CASE_SI
CASE_SO
CASE_SCR_STATE
CASE_SCS0_STATE
CASE_SCS1_STATE
CASE_SCS2_STATE
CASE_SCS3_STATE
CASE_ESC_IGNORE
CASE_ESC_DIGIT
CASE_ESC_SEMI
CASE_DEC_STATE
CASE_ICH
CASE_CUU
CASE_CUD
CASE_CUF
CASE_CUB
CASE_CUP
CASE_ED
CASE_EL
CASE_IL
CASE_DL
CASE_DCH
CASE_DA1
CASE_TRACK_MOUSE
CASE_TBC
CASE_SET
CASE_RST
CASE_SGR
CASE_CPR
CASE_DECSTBM
CASE_DECREQTPARM
CASE_DECSET
CASE_DECRST
CASE_DECALN
CASE_GSETS
CASE_DECSC
CASE_DECRC
CASE_DECKPAM
CASE_DECKPNM
CASE_IND
CASE_NEL
CASE_HTS
CASE_RI
CASE_SS2
CASE_SS3
CASE_CSI_STATE
CASE_OSC
CASE_RIS
CASE_LS2
CASE_LS3
CASE_LS3R
CASE_LS2R
CASE_LS1R
CASE_PRINT
CASE_XTERM_SAVE
CASE_XTERM_RESTORE
CASE_XTERM_TITLE
CASE_DECID
CASE_HP_MEM_LOCK
CASE_HP_MEM_UNLOCK
CASE_HP_BUGGY_LL
CASE_HPA
CASE_VPA
CASE_XTERM_WINOPS
CASE_ECH
CASE_CHT
CASE_CPL
CASE_CNL
CASE_CBT
CASE_SU
CASE_SD
CASE_S7C1T
CASE_S8C1T
CASE_ESC_SP_STATE
CASE_ENQ
CASE_DECSCL
CASE_DECSCA
CASE_DECSED
CASE_DECSEL
CASE_DCS
CASE_PM
CASE_SOS
CASE_ST
CASE_APC
CASE_EPA
CASE_SPA
CASE_CSI_QUOTE_STATE
CASE_DSR
CASE_ANSI_LEVEL_1
CASE_ANSI_LEVEL_2
CASE_ANSI_LEVEL_3
CASE_MC
CASE_DEC2_STATE
CASE_DA2
CASE_DEC3_STATE
CASE_DECRPTUI
CASE_VT52_CUP
CASE_REP
CASE_CSI_EX_STATE
CASE_DECSTR
CASE_DECDHL
CASE_DECSWL
CASE_DECDWL
CASE_DEC_MC
CASE_ESC_PERCENT
CASE_UTF8
CASE_CSI_TICK_STATE
CASE_DECELR
CASE_DECRQLP
CASE_DECEFR
CASE_DECSLE
CASE_CSI_IGNORE
CASE_VT52_IGNORE
CASE_VT52_FINISH
CASE_CSI_DOLLAR_STATE
CASE_DECCRA
CASE_DECERA
CASE_DECFRA
CASE_DECSERA
CASE_DECSACE
CASE_DECCARA
CASE_DECRARA
CASE_CSI_STAR_STATE
CASE_SET_MOD_FKEYS
CASE_SET_MOD_FKEYS0
CASE_HIDE_POINTER
